/*
 * Copyright 2000-2007 Sun Microsystems, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 */

package sun.java2d.xr;

import sun.awt.SunToolkit;
import sun.java2d.loops.*;
import sun.java2d.pipe.Region;
import sun.java2d.SurfaceData;
import java.awt.*;
import java.awt.geom.*;
import java.awt.image.*;

/**
 * XRPMBlitLoops
 * 
 * This class accelerates Blits between two surfaces of types *PM.s
 */
public class XRPMBlitLoops {
    
    public static void register() {
	GraphicsPrimitive[] primitives = {

	new X11PMBlit(XRSurfaceData.IntRgbX11, XRSurfaceData.IntRgbX11), new X11PMBlit(XRSurfaceData.IntRgbX11, XRSurfaceData.IntArgbX11),
		new X11PMBlit(XRSurfaceData.IntArgbX11, XRSurfaceData.IntRgbX11), new X11PMBlit(XRSurfaceData.IntArgbX11, XRSurfaceData.IntArgbX11),

		new X11PMScaledBlit(XRSurfaceData.IntRgbX11, XRSurfaceData.IntRgbX11),
		new X11PMScaledBlit(XRSurfaceData.IntRgbX11, XRSurfaceData.IntArgbX11),
		new X11PMScaledBlit(XRSurfaceData.IntArgbX11, XRSurfaceData.IntRgbX11),
		new X11PMScaledBlit(XRSurfaceData.IntArgbX11, XRSurfaceData.IntArgbX11),

		new X11PMTransformedBlit(XRSurfaceData.IntRgbX11, XRSurfaceData.IntRgbX11),
		new X11PMTransformedBlit(XRSurfaceData.IntRgbX11, XRSurfaceData.IntArgbX11),
		new X11PMTransformedBlit(XRSurfaceData.IntArgbX11, XRSurfaceData.IntRgbX11),
		new X11PMTransformedBlit(XRSurfaceData.IntArgbX11, XRSurfaceData.IntArgbX11), };
	GraphicsPrimitiveMgr.register(primitives);
    }

    public static native void nativeRenderBlit(long srcData, long dstData, int sx, int sy, int dx, int dy, int w, int h);

    public static native void nativeTransformedRenderBlit(long srcData, long dstData, int sx, int sy, int dx, int dy, int w, int h, int m00, int m01,
	    int m02, int m10, int m11, int m12, int maskWidth, int maskHeight, int lastMaskWidth, int lastMaskHeight);
}

class X11PMBlit extends Blit {
    public X11PMBlit(SurfaceType srcType, SurfaceType dstType) {
	super(srcType, CompositeType.AnyAlpha, dstType);
    }

    public void Blit(SurfaceData src, SurfaceData dst, Composite comp, Region clip, int sx, int sy, int dx, int dy, int w, int h) {
	try {
	    SunToolkit.awtLock();

	    XRSurfaceData x11sdDst = (XRSurfaceData) dst;
	    x11sdDst.validate(clip, comp, null, null, null, 0);
	    XRSurfaceData x11sdSrc = (XRSurfaceData) src;

	    x11sdSrc.validateAsSource(null, XRUtils.RepeatNone, XRUtils.FAST);

	    XRPMBlitLoops.nativeRenderBlit(src.getNativeOps(), dst.getNativeOps(), sx, sy, dx, dy, w, h);
	} finally {
	    SunToolkit.awtUnlock();
	}
    }
}

class X11PMScaledBlit extends ScaledBlit {
    public X11PMScaledBlit(SurfaceType srcType, SurfaceType dstType) {
	super(srcType, CompositeType.AnyAlpha, dstType);
    }
    
    /*
     * TODO:  This breaks scales with non-integer coordinates!
     */
    public void Scale(SurfaceData src, SurfaceData dst, Composite comp, Region clip, int sx1, int sy1, int sx2, int sy2, double dx1, double dy1,
	    double dx2, double dy2) {
	try {
	    SunToolkit.awtLock();
	    
//	    int width = sx2-sx1;
//	    int height = sy2-sy1;
//	    double xScale = (dx2 - dx1) / width;
//	    double yScale = (dy2 - dy1) / height;
//	    AffineTransform scaleTransform = AffineTransform.getTranslateInstance(sx1, sy1);
//	    scaleTransform.scale(xScale, yScale);
//	   // scaleTransform.translate(sx1, sy1);
//	    
//	    /*
//             *    static void IsoBlit(SurfaceData src, SurfaceData dst, Composite comp, Region clip, AffineTransform xform, int hint, 
//             *    int srcx, int srcy, int dstx, int dsty, int width, int height)
//	     */
//	    XRPMBlitLoops.IsoBlit(src, dst, comp, clip, scaleTransform, AffineTransformOp.TYPE_NEAREST_NEIGHBOR, 0, 0, (int)(dx1/xScale), (int)(dy1/yScale), width, height);
////
	    XRSurfaceData x11sdDst = (XRSurfaceData) dst;
	    x11sdDst.validate(clip, comp, null, null, null, 0);
	    XRSurfaceData x11sdSrc = (XRSurfaceData) src;
	    
	    double xScale = (dx2 - dx1) / (sx2 - sx1);
	    double yScale = (dy2 - dy1) / (sy2 - sy1);

	    sx1 *= xScale;
	    sx2 *= xScale;
	    sy1 *= yScale;
	    sy2 *= yScale;

	    AffineTransform xForm = AffineTransform.getScaleInstance(1 / xScale, 1 / yScale);

	    x11sdSrc.validateAsSource(xForm, XRUtils.RepeatNone, XRUtils.FAST);
	    
	    XRPMBlitLoops.nativeRenderBlit(src.getNativeOps(), dst.getNativeOps(), (int) sx1, (int) sy1, (int) dx1, (int) dy1, (int) (dx2 - dx1),
		    (int) (dy2 - dy1));

	} finally {
	    SunToolkit.awtUnlock();
	}
    }
}

/**
 * Called also if scale+transform is set
 * 
 * @author Clemens Eisserer
 */
class X11PMTransformedBlit extends TransformBlit {
    
    static int lastMaskWidth = 0, lastMaskHeight = 0;
    
    public X11PMTransformedBlit(SurfaceType srcType, SurfaceType dstType) {
	super(srcType, CompositeType.AnyAlpha, dstType);
    }

    /**
     * Possible optimizations: 
     * - Nearest interpolation -> ExtraAlpha mit 1x1 maske durchführen
     * - Mehere Maskengroessen anlegen, und skalieren
     * - Ueberpruefen ob aktueller Maskeninhalt mit skalierung passen wuerde, bzw. nur so viel aendern bis passt (wenn bereich < maske)
     * - Aendering entweder mit 2 filLRects, oder nur 1 fillrect je nach groesse
     */
    
    public void Transform(SurfaceData src, SurfaceData dst, Composite comp, Region clip, AffineTransform xform, int hint, int srcx, int srcy,
	    int dstx, int dsty, int width, int height) {
	try {
	    SunToolkit.awtLock();
	    int xrInterpolationType = XRUtils.ATransOpToXRQuality(hint);

	    XRSurfaceData x11sdDst = (XRSurfaceData) dst;
	    x11sdDst.validate(clip, comp, null, null, null, 0);
	    XRSurfaceData x11sdSrc = (XRSurfaceData) src;

	    Rectangle2D.Float rect = new Rectangle2D.Float(dstx, dsty, width, height);
	    Shape shp = xform.createTransformedShape(rect);
	    Rectangle bounds = shp.getBounds();

	    AffineTransform trx = AffineTransform.getTranslateInstance((-bounds.x), (-bounds.y));
	    trx.concatenate(xform);
	    AffineTransform maskTX = (AffineTransform) trx.clone();

	    trx.translate(-srcx, -srcy);

	    try {
		trx.invert();
	    } catch (NoninvertibleTransformException ex) {
		trx.setToIdentity();
		System.err.println("Reseted to identity!");
	    }

	    boolean omitMask = isMaskOmittable(trx, comp, xrInterpolationType);

	    if (!omitMask) {
		int maskWidth = Math.max(width / 8, 1);
		int maskHeight = Math.max(height / 8, 1);
		maskTX.scale(((double) width) / maskWidth, ((double) height) / maskHeight);

		try {
		    maskTX.invert();
		} catch (NoninvertibleTransformException ex) {
		    maskTX.setToIdentity();
		    System.err.println("Reseted to identity!");
		}

		x11sdSrc.validateAsSource(trx, XRUtils.RepeatPad, xrInterpolationType);

		XRPMBlitLoops.nativeTransformedRenderBlit(src.getNativeOps(), dst.getNativeOps(), 0, 0, bounds.x, bounds.y, bounds.width,
			bounds.height, XRUtils.XDoubleToFixed(maskTX.getScaleX()), XRUtils.XDoubleToFixed(maskTX.getShearX()), XRUtils
				.XDoubleToFixed(maskTX.getTranslateX()), XRUtils.XDoubleToFixed(maskTX.getShearY()), XRUtils.XDoubleToFixed(maskTX
				.getScaleY()), XRUtils.XDoubleToFixed(maskTX.getTranslateY()), maskWidth, maskHeight, lastMaskWidth, lastMaskHeight);

		lastMaskWidth = maskWidth;
		lastMaskHeight = maskHeight;
	    } else {
		int repeat = xrInterpolationType <= XRUtils.FAST ? XRUtils.RepeatNone : XRUtils.RepeatPad;
		
		x11sdSrc.validateAsSource(trx, repeat, xrInterpolationType);
		XRPMBlitLoops.nativeTransformedRenderBlit(src.getNativeOps(), dst.getNativeOps(), 0, 0, bounds.x, bounds.y, bounds.width,
			bounds.height, 0, 0, 0, 0, 0, 0, -1, -1, 0, 0);
	    }
	} finally {
	    SunToolkit.awtUnlock();
	}
    }
    
    protected static boolean isMaskOmittable(AffineTransform trx, Composite comp, int interpolation) {
	return (interpolation <= XRUtils.FAST || trx.getTranslateX() == (int) trx.getTranslateX() /*
												     * If
												     * translate
												     * is
												     * integer only
												     */
		&& trx.getTranslateY() == (int) trx.getTranslateY() && (trx.getShearX() == 0 && trx.getShearY() == 0 // Only
															// "90°"
															// rotation
		|| trx.getShearX() == -trx.getShearY())) && ((AlphaComposite) comp).getAlpha() == 1.0f; // No
													// ExtraAlpha!=1
    }
}